from typing import Dict, Any

import sshtunnel
import urllib
import urllib.request
import urllib.error

from paramiko import ProxyCommand

from cloudtik.core._private.cluster.cluster_config import _load_cluster_config
from cloudtik.core._private.utils import get_cluster_head_ip, is_use_internal_ip

REST_ENDPOINT_URL_FORMAT = "http://{}:{}/{}"
REST_REQUEST_TIMEOUT = 60


def request_rest_to_head(
        cluster_config_file: str, cluster_name: str, endpoint: str, rest_api_port: int,
        on_head: bool = False):
    config = _load_cluster_config(cluster_config_file, cluster_name)
    return _request_rest_to_head(
        config=config, endpoint=endpoint, rest_api_port=rest_api_port,
        on_head=on_head)


def _request_rest_to_head(
        config: Dict[str, Any], endpoint: str, rest_api_port: int,
        on_head: bool = False):
    head_node_ip = get_cluster_head_ip(config, False)
    if on_head or is_use_internal_ip(config):
        return request_rest_direct(
            rest_api_ip=head_node_ip, rest_api_port=rest_api_port, endpoint=endpoint)
    else:
        head_public_ip = get_cluster_head_ip(config, True)
        return request_rest_to_server(
            config=config, server_ip=head_public_ip,
            rest_api_ip=head_node_ip, rest_api_port=rest_api_port, endpoint=endpoint)


def ssh_proxy_wrapper(ssh_proxy_command, server_ip, ssh_port, ssh_user):
    if ssh_proxy_command is None:
        return None
    ssh_proxy_command = ssh_proxy_command.replace("%h", server_ip)
    ssh_proxy_command = ssh_proxy_command.replace("%p", str(ssh_port))
    ssh_proxy_command = ssh_proxy_command.replace("%r", str(ssh_user))
    ssh_proxy = ProxyCommand(ssh_proxy_command)
    return ssh_proxy


def request_rest_to_server(
        config, server_ip, rest_api_ip: str, rest_api_port: int, endpoint: str):
    auth_config = config.get("auth", {})
    ssh_proxy_command = auth_config.get("ssh_proxy_command", None)
    ssh_private_key = auth_config.get("ssh_private_key", None)
    ssh_user = auth_config["ssh_user"]
    ssh_port = auth_config.get("ssh_port", 22)
    ssh_proxy = ssh_proxy_wrapper(ssh_proxy_command, server_ip, ssh_port, ssh_user)

    with sshtunnel.open_tunnel(
            (server_ip, int(ssh_port)),
            ssh_username=ssh_user,
            ssh_pkey=ssh_private_key,
            ssh_proxy=ssh_proxy,
            remote_bind_address=(rest_api_ip, rest_api_port)
    ) as tunnel:
        return request_rest_direct("127.0.0.1", tunnel.local_bind_port, endpoint)


def request_rest_direct(rest_api_ip: str, rest_api_port: int, endpoint: str):
    endpoint_url = REST_ENDPOINT_URL_FORMAT.format(
        rest_api_ip, rest_api_port, endpoint)

    # sine we have use 127.0.0.1, disable all proxy on 127.0.0.1
    proxy_support = urllib.request.ProxyHandler({"no": "127.0.0.1"})
    opener = urllib.request.build_opener(proxy_support)
    urllib.request.install_opener(opener)

    response = urllib.request.urlopen(endpoint_url, timeout=REST_REQUEST_TIMEOUT)
    return response.read()


def request_tunnel_to_server(
        config, server_ip, remote_ip: str, remote_port: int,
        request_fn, args=(), kwargs={}):
    auth_config = config.get("auth", {})
    ssh_proxy_command = auth_config.get("ssh_proxy_command", None)
    ssh_private_key = auth_config.get("ssh_private_key", None)
    ssh_user = auth_config["ssh_user"]
    ssh_port = auth_config.get("ssh_port", 22)
    ssh_proxy = ssh_proxy_wrapper(ssh_proxy_command, server_ip, ssh_port, ssh_user)
    with sshtunnel.open_tunnel(
            server_ip,
            ssh_username=ssh_user,
            ssh_port=ssh_port,
            ssh_pkey=ssh_private_key,
            ssh_proxy=ssh_proxy,
            remote_bind_address=(remote_ip, remote_port)
    ) as tunnel:
        return request_fn("127.0.0.1", tunnel.local_bind_port, *args, **kwargs)


def request_tunnel_to_head(
        config: Dict[str, Any], target_port: int,
        request_fn, args=(), kwargs={},
        on_head: bool = False):
    head_node_ip = get_cluster_head_ip(config, False)
    if on_head:
        return request_fn(head_node_ip, target_port, *args, **kwargs)
    else:
        head_public_ip = get_cluster_head_ip(config, True)
        return request_tunnel_to_server(
            config=config, server_ip=head_public_ip,
            remote_ip=head_node_ip, remote_port=target_port,
            request_fn=request_fn, args=args, kwargs=kwargs)
