import logging
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple

from cloudtik.core._private.annotations import DeveloperAPI
from cloudtik.core._private.call_context import CallContext
from cloudtik.core._private.cli_logger import cli_logger
from cloudtik.core.command_executor import CommandExecutor
from cloudtik.core._private.command_executor.ssh_command_executor import \
    SSHCommandExecutor
from cloudtik.core._private.command_executor.docker_command_executor import \
    DockerCommandExecutor

logger = logging.getLogger(__name__)


@DeveloperAPI
class NodeLaunchException(Exception):
    """A structured exception that can be thrown by a node provider during a
    `create_node` call to pass additional information for observability.
    """

    def __init__(
        self,
        category: str,
        description: str,
        src_exc_info: Optional[Tuple[Any, Any, Any]],  # The
    ):
        """Args:
        category: A short (<20 chars) label for the error.
        description: A longer, human-readable description of the error.
        src_exc_info: The source exception info if applicable. This is a
              tuple of (type, exception, traceback) as returned by
              sys.exc_info()

        """
        super().__init__(f"Node Launch Exception ({category}): {description}")
        self.category = category
        self.description = description
        self.src_exc_info = src_exc_info

    def __reduce__(self):
        # NOTE: Since tracebacks can't be pickled, we'll drop the optional
        # traceback if we have to serialize this object.
        return (
            self.__class__,
            (self.category, self.description, None),
        )


@DeveloperAPI
class NodeProvider:
    """Interface for getting and returning nodes from a Cloud.

    **Important**: This is an INTERNAL API that is only exposed for the purpose
    of implementing custom node providers. It is not allowed to call into
    NodeProvider methods from any package outside, only to
    define new implementations of NodeProvider for use with the "external" node
    provider option.

    NodeProviders are namespaced by the `cluster_name` parameter; they only
    operate on nodes within that namespace.

    Nodes may be in one of three states: {pending, running, terminated}. Nodes
    appear immediately once started by `create_node`, and transition
    immediately to terminated when `terminate_node` is called.
    """

    def __init__(self, provider_config: Dict[str, Any],
                 cluster_name: str) -> None:
        self.provider_config = provider_config
        self.cluster_name = cluster_name
        self._internal_ip_cache: Dict[str, str] = {}
        self._external_ip_cache: Dict[str, str] = {}

    def non_terminated_nodes(self, tag_filters: Dict[str, str]) -> List[str]:
        """Return a list of node ids filtered by the specified tags dict.

        This list must not include terminated nodes. For performance reasons,
        providers are allowed to cache the result of a call to
        non_terminated_nodes() to serve single-node queries
        (e.g. is_running(node_id)). This means that non_terminate_nodes() must
        be called again to refresh results.

        Examples:
            >>> provider.non_terminated_nodes({CLOUDTIK_TAG_NODE_KIND: "worker"})
            ["node-1", "node-2"]
        """
        raise NotImplementedError

    def is_running(self, node_id: str) -> bool:
        """Return whether the specified node is running."""
        raise NotImplementedError

    def is_terminated(self, node_id: str) -> bool:
        """Return whether the specified node is terminated."""
        raise NotImplementedError

    def node_tags(self, node_id: str) -> Dict[str, str]:
        """Returns the tags of the given node (string dict)."""
        raise NotImplementedError

    def external_ip(self, node_id: str) -> str:
        """Returns the external ip of the given node."""
        raise NotImplementedError

    def internal_ip(self, node_id: str) -> str:
        """Returns the internal ip of the given node."""
        raise NotImplementedError

    def get_node_id(self, ip_address: str,
                    use_internal_ip: bool = False) -> str:
        """Returns the node_id given an IP address.

        Assumes ip-address is unique per node.

        Args:
            ip_address (str): Address of node.
            use_internal_ip (bool): Whether the ip address is
                public or private.

        Raises:
            ValueError if not found.
        """

        def find_node_id():
            if use_internal_ip:
                return self._internal_ip_cache.get(ip_address)
            else:
                return self._external_ip_cache.get(ip_address)

        if not find_node_id():
            all_nodes = self.non_terminated_nodes({})
            get_ip = self.internal_ip if use_internal_ip else self.external_ip
            ip_cache = self._internal_ip_cache if use_internal_ip else self._external_ip_cache
            for node_id in all_nodes:
                ip_cache[get_ip(node_id)] = node_id

        if not find_node_id():
            if use_internal_ip:
                known_msg = (
                    f"Worker internal IPs: {list(self._internal_ip_cache)}")
            else:
                known_msg = (
                    f"Worker external IP: {list(self._external_ip_cache)}")
            raise ValueError(f"ip {ip_address} not found. " + known_msg)

        return find_node_id()

    def create_node(self, node_config: Dict[str, Any], tags: Dict[str, str],
                    count: int) -> Optional[Dict[str, Any]]:
        """Creates a number of nodes within the namespace.

        Optionally returns a mapping from created node ids to node metadata.
        Optionally may throw a NodeLaunchException with detailed information of the launch failure.
        """
        raise NotImplementedError

    def create_node_with_resources(
            self, node_config: Dict[str, Any], tags: Dict[str, str],
            count: int,
            resources: Dict[str, float]) -> Optional[Dict[str, Any]]:
        """Create nodes with a given resource config.

        This is the method actually called by the cluster scaler. Prefer to
        implement this when possible directly, otherwise it delegates to the
        create_node() implementation.

        Optionally may throw a NodeLaunchException with detailed information of the launch failure.
        """
        return self.create_node(node_config, tags, count)

    def set_node_tags(self, node_id: str, tags: Dict[str, str]) -> None:
        """Sets the tag values (string dict) for the specified node."""
        raise NotImplementedError

    def terminate_node(self, node_id: str) -> Optional[Dict[str, Any]]:
        """Terminates the specified node.

        Optionally return a mapping from deleted node ids to node
        metadata.
        """
        raise NotImplementedError

    def terminate_nodes(self, node_ids: List[str]) -> Optional[Dict[str, Any]]:
        """Terminates a set of nodes.

        May be overridden with a batch method, which optionally may return a
        mapping from deleted node ids to node metadata.
        """
        for node_id in node_ids:
            cli_logger.verbose("NodeProvider: "
                               "{}: Terminating node".format(node_id))
            self.terminate_node(node_id)
        return None

    @property
    def max_terminate_nodes(self) -> Optional[int]:
        """The maximum number of nodes which can be terminated in one single
        API request. By default, this is "None", which means that the node
        provider's underlying API allows infinite requests to be terminated
        with one request.

        For example, AWS only allows 1000 nodes to be terminated
        at once; to terminate more, we must issue multiple separate API
        requests. If the limit is infinity, then simply set this to None.

        This may be overridden. The value may be useful when overriding the
        "terminate_nodes" method.
        """
        return None

    def get_command_executor(self,
                             call_context: CallContext,
                             log_prefix: str,
                             node_id: str,
                             auth_config: Dict[str, Any],
                             cluster_name: str,
                             process_runner: ModuleType,
                             use_internal_ip: bool,
                             docker_config: Optional[Dict[str, Any]] = None
                             ) -> CommandExecutor:
        """Returns the CommandRunner class used to perform SSH commands.

        Args:
        call_context: The call context.
        log_prefix(str): stores "NodeUpdater: {}: ".format(<node_id>). Used
            to print progress in the CommandRunner.
        node_id(str): the node ID.
        auth_config(dict): the authentication configs from the cluster config
            yaml file.
        cluster_name(str): the name of the cluster.
        process_runner(module): the module to use to run the commands
            in the CommandRunner. E.g., subprocess.
        use_internal_ip(bool): whether the node_id belongs to an internal ip
            or external ip.
        docker_config(dict): If set, the docker information of the docker
            container that commands should be run on.
        """
        common_args = {
            "log_prefix": log_prefix,
            "auth_config": auth_config,
            "cluster_name": cluster_name,
            "process_runner": process_runner,
            "use_internal_ip": use_internal_ip,
            "provider": self,
            "node_id": node_id,
        }
        if docker_config and docker_config.get("enabled", False):
            return DockerCommandExecutor(
                call_context, docker_config, True, **common_args)
        else:
            return SSHCommandExecutor(call_context, **common_args)

    def prepare_config_for_head(
            self, cluster_config: Dict[str, Any], remote_config: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a new remote cluster config with custom configs for head node.
        The cluster config may also be updated for setting up the head"""
        return remote_config

    def cleanup_cluster(
            self, cluster_config: Dict[str, Any], deep: bool = False):
        """Cleanup the cluster by deleting additional resources other than the nodes.
        If deep flag is true, do a deep clean up all the resources
        """
        pass

    def get_node_info(self, node_id: str) -> Dict[str, str]:
        """Return the node detail information  of a instance by instance id .
        Examples:
            >>> provider.get_node_info("i-01206b67a0bf03dee")
            {
                "node_id": "i-01206b67a0bf03dee",
                "instance_type": "m5.large",
                "private_ip": "172.31.53.194",
                "public_ip": "44.242.153.247",
                "instance_status": "running",
                "cloudtik-launch-config": "be65513f591bd85758abb1e3dc4e8c51e26bd69a",
                "cloudtik-user-node-type": "worker.default",
                "cloudtik-node-kind": "worker",
                "cloudtik-node-status": "up-to-date",
                "Name": "cloudtik-example-worker",
                "cloudtik-cluster-name": "example",
                "cloudtik-runtime-config": "63611285826253d8672bbc3a0bad3405ae8891d7"
            }
        """
        raise NotImplementedError

    def with_environment_variables(self, node_type_config: Dict[str, Any], node_id: str):
        """Export necessary environment variables for running node commands"""
        raise NotImplementedError

    def get_default_cloud_storage(self):
        """Return the managed cloud storage if configured."""
        return None

    def get_default_cloud_database(self):
        """Return the configured cloud database if configured."""
        return None

    @staticmethod
    def prepare_config(cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Prepare the necessary configs for user before merge with system defaults
        This is the first process after the config is loaded.
        """
        return cluster_config

    @staticmethod
    def post_prepare(
            cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Fills out missing fields after the user config is merged with defaults
        This happens after prepare_config is done.
        """
        return cluster_config

    @staticmethod
    def validate_config(
            provider_config: Dict[str, Any]) -> None:
        """Check the provider configuration validation.
        This happens after post_prepare is done and before bootstrap_config
        """
        return None

    @staticmethod
    def bootstrap_config(cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Bootstraps the cluster config by adding env defaults if needed.
        This happens after validate_config is done.
        """
        return cluster_config

    @staticmethod
    def bootstrap_config_for_api(cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Bootstraps the cluster config for node provider api access.
        This happens after validate_config is done.
        """
        return cluster_config

    @staticmethod
    def verify_config(
            provider_config: Dict[str, Any]) -> None:
        """Verify provider configuration. Verification usually means to check it is working.
        This happens after bootstrap_config is done.
        """
        return None
