import copy
import sys
import threading
from collections import defaultdict, OrderedDict
import logging
import time
from typing import Any, Dict, List

import botocore

from cloudtik.core._private.utils import _is_permanent_data_volumes
from cloudtik.core.node_provider import NodeProvider, NodeLaunchException
from cloudtik.core.tags import CLOUDTIK_TAG_CLUSTER_NAME, CLOUDTIK_TAG_NODE_NAME, \
    CLOUDTIK_TAG_LAUNCH_CONFIG, CLOUDTIK_TAG_NODE_KIND, CLOUDTIK_TAG_USER_NODE_TYPE

from cloudtik.core._private.log_timer import LogTimer
from cloudtik.core._private.cli_logger import cli_logger, cf

from cloudtik.providers._private.aws.config import verify_s3_storage, bootstrap_aws, post_prepare_aws, \
    with_aws_environment_variables, delete_cluster_volumes, create_volumes_for_node, attach_volumes_for_node, \
    rollback_volumes_for_node
from cloudtik.providers._private.aws.utils import boto_exception_handler, \
    get_boto_error_code, BOTO_MAX_RETRIES, BOTO_CREATE_MAX_RETRIES, \
    _get_node_info, make_ec2_resource, get_aws_s3_storage_config, get_default_aws_cloud_storage, \
    get_default_aws_cloud_database, get_aws_credentials, clear_aws_credentials
from cloudtik.providers._private.utils import validate_config_dict

logger = logging.getLogger(__name__)

TAG_BATCH_DELAY = 1


def to_aws_format(tags):
    """Convert the node name tag to the AWS-specific 'Name' tag."""

    if CLOUDTIK_TAG_NODE_NAME in tags:
        tags["Name"] = tags[CLOUDTIK_TAG_NODE_NAME]
        del tags[CLOUDTIK_TAG_NODE_NAME]
    return tags


def from_aws_format(tags):
    """Convert the AWS-specific 'Name' tag to the node name tag."""

    if "Name" in tags:
        tags[CLOUDTIK_TAG_NODE_NAME] = tags["Name"]
        del tags["Name"]
    return tags


class AWSNodeProvider(NodeProvider):
    """
    Computing nodes provider for AWS.
    All the operations use instance.id as the unique node id.
    Which is the same as shown in the node-id of status command.
    The node id is auto-generated by cloud provider.
    And CLOUDTIK_TAG_NODE_NAME tag becomes the Name tag of the instance.
    """
    max_terminate_nodes = 1000

    def __init__(self, provider_config, cluster_name):
        NodeProvider.__init__(self, provider_config, cluster_name)
        self.cache_stopped_nodes = provider_config.get("cache_stopped_nodes",
                                                       False)
        aws_credentials = get_aws_credentials(provider_config)

        self.ec2 = make_ec2_resource(
            region=provider_config["region"],
            max_retries=BOTO_MAX_RETRIES,
            aws_credentials=aws_credentials)
        self.ec2_fail_fast = make_ec2_resource(
            region=provider_config["region"],
            max_retries=0,
            aws_credentials=aws_credentials)

        # Tags that we believe to actually be on EC2.
        self.tag_cache = {}
        # Tags that we will soon upload.
        self.tag_cache_pending = defaultdict(dict)
        # Number of threads waiting for a batched tag update.
        self.batch_thread_count = 0
        self.batch_update_done = threading.Event()
        self.batch_update_done.set()
        self.ready_for_new_batch = threading.Event()
        self.ready_for_new_batch.set()
        self.tag_cache_lock = threading.Lock()
        self.count_lock = threading.Lock()

        # Cache of node objects from the last nodes() call. This avoids
        # excessive DescribeInstances requests.
        self.cached_nodes = {}

    def with_environment_variables(self, node_type_config: Dict[str, Any], node_id: str):
        return with_aws_environment_variables(self.provider_config, node_type_config, node_id)

    def non_terminated_nodes(self, tag_filters):
        # Note that these filters are acceptable because they are set on
        #       node initialization, and so can never be sitting in the cache.
        tag_filters = to_aws_format(tag_filters)
        filters = [
            {
                "Name": "instance-state-name",
                "Values": ["pending", "running"],
            },
            {
                "Name": "tag:{}".format(CLOUDTIK_TAG_CLUSTER_NAME),
                "Values": [self.cluster_name],
            },
        ]
        for k, v in tag_filters.items():
            filters.append({
                "Name": "tag:{}".format(k),
                "Values": [v],
            })

        with boto_exception_handler(
                "Failed to fetch running instances from AWS."):
            nodes = list(self.ec2.instances.filter(Filters=filters))

        # Populate the tag cache with initial information if necessary
        for node in nodes:
            if node.id in self.tag_cache:
                continue

            self.tag_cache[node.id] = from_aws_format(
                {x["Key"]: x["Value"]
                 for x in node.tags})

        self.cached_nodes = {node.id: node for node in nodes}
        return [node.id for node in nodes]

    def get_node_info(self, node_id):
        node = self._get_cached_node(node_id)
        return _get_node_info(node)

    def is_running(self, node_id):
        node = self._get_cached_node(node_id)
        return node.state["Name"] == "running"

    def is_terminated(self, node_id):
        node = self._get_cached_node(node_id)
        state = node.state["Name"]
        return state not in ["running", "pending"]

    def node_tags(self, node_id):
        with self.tag_cache_lock:
            d1 = self.tag_cache[node_id]
            d2 = self.tag_cache_pending.get(node_id, {})
            return dict(d1, **d2)

    def external_ip(self, node_id):
        node = self._get_cached_node(node_id)

        if node.public_ip_address is None:
            node = self._get_node(node_id)

        return node.public_ip_address

    def internal_ip(self, node_id):
        node = self._get_cached_node(node_id)

        if node.private_ip_address is None:
            node = self._get_node(node_id)

        return node.private_ip_address

    def set_node_tags(self, node_id, tags):
        is_batching_thread = False
        with self.tag_cache_lock:
            if not self.tag_cache_pending:
                is_batching_thread = True
                # Wait for threads in the last batch to exit
                self.ready_for_new_batch.wait()
                self.ready_for_new_batch.clear()
                self.batch_update_done.clear()
            self.tag_cache_pending[node_id].update(tags)

        if is_batching_thread:
            time.sleep(TAG_BATCH_DELAY)
            with self.tag_cache_lock:
                self._update_node_tags()
                self.batch_update_done.set()

        with self.count_lock:
            self.batch_thread_count += 1
        self.batch_update_done.wait()

        with self.count_lock:
            self.batch_thread_count -= 1
            if self.batch_thread_count == 0:
                self.ready_for_new_batch.set()

    def _update_node_tags(self):
        batch_updates = defaultdict(list)

        for node_id, tags in self.tag_cache_pending.items():
            for x in tags.items():
                batch_updates[x].append(node_id)
            self.tag_cache[node_id].update(tags)

        self.tag_cache_pending = defaultdict(dict)

        self._create_tags(batch_updates)

    def _create_tags(self, batch_updates):
        for (k, v), node_ids in batch_updates.items():
            m = "Set tag {}={} on {}".format(k, v, node_ids)
            with LogTimer("AWSNodeProvider: {}".format(m)):
                if k == CLOUDTIK_TAG_NODE_NAME:
                    k = "Name"
                self.ec2.meta.client.create_tags(
                    Resources=node_ids,
                    Tags=[{
                        "Key": k,
                        "Value": v
                    }],
                )

    def create_node(self, node_config, tags, count) -> Dict[str, Any]:
        """Creates instances.

        Returns dict mapping instance id to ec2.Instance object for the created
        instances.
        """
        # sort tags by key to support deterministic unit test stubbing
        tags = OrderedDict(sorted(copy.deepcopy(tags).items()))

        reused_nodes_dict = {}
        # Try to reuse previously stopped nodes with compatible configs
        if self.cache_stopped_nodes:
            # TODO(ekl) this is breaking the abstraction boundary a little by
            # peeking into the tag set.
            filters = [
                {
                    "Name": "instance-state-name",
                    "Values": ["stopped", "stopping"],
                },
                {
                    "Name": "tag:{}".format(CLOUDTIK_TAG_CLUSTER_NAME),
                    "Values": [self.cluster_name],
                },
                {
                    "Name": "tag:{}".format(CLOUDTIK_TAG_NODE_KIND),
                    "Values": [tags[CLOUDTIK_TAG_NODE_KIND]],
                },
                {
                    "Name": "tag:{}".format(CLOUDTIK_TAG_LAUNCH_CONFIG),
                    "Values": [tags[CLOUDTIK_TAG_LAUNCH_CONFIG]],
                },
            ]
            # This tag may not always be present.
            if CLOUDTIK_TAG_USER_NODE_TYPE in tags:
                filters.append({
                    "Name": "tag:{}".format(CLOUDTIK_TAG_USER_NODE_TYPE),
                    "Values": [tags[CLOUDTIK_TAG_USER_NODE_TYPE]],
                })

            reuse_nodes = list(
                self.ec2.instances.filter(Filters=filters))[:count]
            reuse_node_ids = [n.id for n in reuse_nodes]
            reused_nodes_dict = {n.id: n for n in reuse_nodes}
            if reuse_nodes:
                cli_logger.print(
                    # todo: handle plural vs singular?
                    "Reusing nodes {}. "
                    "To disable reuse, set `cache_stopped_nodes: False` "
                    "under `provider` in the cluster configuration.",
                    cli_logger.render_list(reuse_node_ids))

                # todo: timed?
                with cli_logger.group("Stopping instances to reuse"):
                    for node in reuse_nodes:
                        self.tag_cache[node.id] = from_aws_format(
                            {x["Key"]: x["Value"]
                             for x in node.tags})
                        if node.state["Name"] == "stopping":
                            cli_logger.print("Waiting for instance {} to stop",
                                             node.id)
                            node.wait_until_stopped()

                self.ec2.meta.client.start_instances(
                    InstanceIds=reuse_node_ids)
                for node_id in reuse_node_ids:
                    self.set_node_tags(node_id, tags)
                count -= len(reuse_node_ids)

        created_nodes_dict = {}
        if count:
            created_nodes_dict = self._create_node(node_config, tags, count)

        all_created_nodes = reused_nodes_dict
        all_created_nodes.update(created_nodes_dict)
        return all_created_nodes

    @staticmethod
    def _merge_tag_specs(tag_specs: List[Dict[str, Any]],
                         user_tag_specs: List[Dict[str, Any]]) -> None:
        """
        Merges user-provided node config tag specifications into a base
        list of node provider tag specifications. The base list of
        node provider tag specs is modified in-place.

        This allows users to add tags and override values of existing
        tags with their own, and only applies to the resource type
        "instance". All other resource types are appended to the list of
        tag specs.

        Args:
            tag_specs (List[Dict[str, Any]]): base node provider tag specs
            user_tag_specs (List[Dict[str, Any]]): user's node config tag specs
        """

        for user_tag_spec in user_tag_specs:
            if user_tag_spec["ResourceType"] == "instance":
                for user_tag in user_tag_spec["Tags"]:
                    exists = False
                    for tag in tag_specs[0]["Tags"]:
                        if user_tag["Key"] == tag["Key"]:
                            exists = True
                            tag["Value"] = user_tag["Value"]
                            break
                    if not exists:
                        tag_specs[0]["Tags"] += [user_tag]
            else:
                tag_specs += [user_tag_spec]

    def _create_node(self, node_config, tags, count):
        created_nodes_dict = {}

        tags = to_aws_format(tags)
        conf = node_config.copy()

        tag_pairs = [{
            "Key": CLOUDTIK_TAG_CLUSTER_NAME,
            "Value": self.cluster_name,
        }]
        for k, v in tags.items():
            tag_pairs.append({
                "Key": k,
                "Value": v,
            })
        tag_specs = [{
            "ResourceType": "instance",
            "Tags": tag_pairs,
        }]
        user_tag_specs = conf.get("TagSpecifications", [])
        AWSNodeProvider._merge_tag_specs(tag_specs, user_tag_specs)

        # SubnetIds is not a real config key: we must resolve to a
        # single SubnetId before invoking the AWS API.
        subnet_ids = conf.pop("SubnetIds")

        # update config with min/max node counts and tag specs
        conf.update({
            "MinCount": 1,
            "MaxCount": count,
            "TagSpecifications": tag_specs
        })

        # Try to always launch in the first listed subnet.
        subnet_idx = 0
        cli_logger_tags = {}

        # provision volumes if needed (for the count will be 1 for this case)
        volumes_for_node = {}
        volume_availability_zone = None
        subnet_zone_mapping = self.provider_config.get(
            "subnet_zone_mapping", {})

        # NOTE: This ensures that we try ALL availability zones before
        # throwing an error.
        max_tries = max(BOTO_CREATE_MAX_RETRIES, len(subnet_ids))
        for attempt in range(1, max_tries + 1):
            try:
                if "NetworkInterfaces" in conf:
                    net_ifs = conf["NetworkInterfaces"]
                    # remove security group IDs previously copied from network
                    # interfaces (create_instances call fails otherwise)
                    conf.pop("SecurityGroupIds", None)
                    cli_logger_tags["network_interfaces"] = str(net_ifs)
                    subnet_id = conf["SubnetId"]
                else:
                    if volume_availability_zone:
                        subnet_id = self._get_subnet_of_zone(
                            subnet_zone_mapping, subnet_ids, volume_availability_zone)
                    else:
                        subnet_id = subnet_ids[subnet_idx % len(subnet_ids)]
                        conf["SubnetId"] = subnet_id
                        cli_logger_tags["subnet_id"] = subnet_id

                # get availability_zone of the subnet
                availability_zone = subnet_zone_mapping.get(subnet_id)
                conf, volume_availability_zone = create_volumes_for_node(
                    self.provider_config, self.ec2, self.cluster_name,
                    node_config, tags, conf, volumes_for_node, availability_zone)
                if volume_availability_zone:
                    # update subnet id to the same zone of volume
                    subnet_id = self._get_subnet_of_zone(
                        subnet_zone_mapping, subnet_ids, volume_availability_zone)
                    conf["SubnetId"] = subnet_id
                    cli_logger_tags["subnet_id"] = subnet_id

                created = self.ec2_fail_fast.create_instances(**conf)
                created_nodes_dict = {n.id: n for n in created}

                # todo: timed?
                # todo: handle plurality?
                with cli_logger.group(
                        "Launched {} nodes", count, _tags=cli_logger_tags):
                    for instance in created:
                        # NOTE(maximsmol): This is needed for mocking
                        # boto3 for tests. This is likely a bug in moto
                        # but AWS docs don't seem to say.
                        # You can patch moto/ec2/responses/instances.py
                        # to fix this (add <stateReason> to EC2_RUN_INSTANCES)

                        # The correct value is technically
                        # {"code": "0", "Message": "pending"}
                        state_reason = instance.state_reason or {
                            "Message": "pending"
                        }

                        cli_logger.print(
                            "Launched instance {}",
                            instance.instance_id,
                            _tags=dict(
                                state=instance.state["Name"],
                                info=state_reason["Message"]))
                break
            except botocore.exceptions.ClientError as exc:
                # Launch failure may be due to instance type availability in
                # the given AZ
                # because the volume is availability zone bounded, try another zone need new volumes
                rollback_volumes_for_node(
                    self.ec2, volumes_for_node)

                error_code = get_boto_error_code(exc)
                if (error_code and
                        error_code == "InsufficientInstanceCapacity" and
                        "InstanceMarketOptions" in conf):
                    # If failed with insufficient capacity, we retry once to request on-demand if it requested spot
                    cli_logger.warning("Retrying request for on-demand instance instead of spot.")
                    conf.pop("InstanceMarketOptions")
                else:
                    if volume_availability_zone:
                        # there are existing volumes, cannot try another zone
                        self._fail_node_creation(
                            exc, "Will not retry other zones because existing volumes.")

                    subnet_idx += 1
                    if attempt == max_tries:
                        self._fail_node_creation(exc, "Max attempts exceeded.")
                    else:
                        cli_logger.warning(
                            "Create instances attempt failed: {}. Retrying...",
                            exc)

        self._attach_node_volumes(
            created_nodes_dict, volumes_for_node)
        return created_nodes_dict

    def _attach_node_volumes(
            self, created_nodes, volumes_for_node):
        try:
            # attach volumes if needed ( if attach failed, terminate it the node)
            attach_volumes_for_node(
                self.ec2, created_nodes, volumes_for_node)
        except botocore.exceptions.ClientError as exc:
            # do clean up
            nodes_to_terminate = list(created_nodes.items())
            for node_id, node in nodes_to_terminate:
                try:
                    # terminate the node
                    node.terminate()
                    created_nodes.pop(node_id, None)
                except Exception as e:
                    logger.warning(
                        "Error terminating node after node creation failure: {}".format(
                            str(e)))
            try:
                # clean up new created volumes
                rollback_volumes_for_node(
                    self.ec2, volumes_for_node)
            except Exception as e:
                logger.warning(
                    "Error cleaning up the volumes after node creation failure: {}".format(
                        str(e)))

            self._fail_node_creation(exc, "Error attaching volumes.")

    @staticmethod
    def _get_subnet_of_zone(subnet_zone_mapping, subnet_ids, availability_zone):
        for subnet_id, zone in subnet_zone_mapping.items():
            if availability_zone == zone and subnet_id in subnet_ids:
                return subnet_id
        raise RuntimeError(
            "Failed to get subnet id of availability zone: {}".format(availability_zone))

    @staticmethod
    def _fail_node_creation(exc, reason):
        try:
            exc = NodeLaunchException(
                category=exc.response["Error"]["Code"],
                description=exc.response["Error"]["Message"],
                src_exc_info=sys.exc_info(),
            )
        except Exception:
            # In theory, all ClientError's we expect to get should
            # have these fields, but just in case we can't parse
            # it, it's fine, just throw the original error.
            logger.warning("Couldn't parse exception.", exc)
            pass
        cli_logger.abort(
            "Failed to launch instances. " + reason,
            exc=exc,
        )

    def terminate_node(self, node_id):
        node = self._get_cached_node(node_id)
        if self.cache_stopped_nodes:
            if node.spot_instance_request_id:
                cli_logger.print(
                    "Terminating instance {} " +
                    cf.dimmed("(cannot stop spot instances, only terminate)"),
                    node_id)  # todo: show node name?
                node.terminate()
            else:
                cli_logger.print("Stopping instance {} " + cf.dimmed(
                    "(to terminate instead, "
                    "set `cache_stopped_nodes: False` "
                    "under `provider` in the cluster configuration)"),
                                 node_id)  # todo: show node name?
                node.stop()
        else:
            node.terminate()

        # TODO: We are leaking the tag cache here. Naively, we would
        # want to just remove the cache entry here, but terminating can be
        # asyncrhonous or error, which would result in a use after free error.
        # If this leak becomes bad, we can garbage collect the tag cache when
        # the node cache is updated.
        pass

    def terminate_nodes(self, node_ids):
        if not node_ids:
            return

        terminate_instances_func = self.ec2.meta.client.terminate_instances
        stop_instances_func = self.ec2.meta.client.stop_instances

        # In some cases, this function stops some nodes, but terminates others.
        # Each of these requires a different EC2 API call. So, we use the
        # "nodes_to_terminate" dict below to keep track of exactly which API
        # call will be used to stop/terminate which set of nodes. The key is
        # the function to use, and the value is the list of nodes to terminate
        # with that function.
        nodes_to_terminate = {
            terminate_instances_func: [],
            stop_instances_func: []
        }

        if self.cache_stopped_nodes:
            spot_ids = []
            on_demand_ids = []

            for node_id in node_ids:
                if self._get_cached_node(node_id).spot_instance_request_id:
                    spot_ids += [node_id]
                else:
                    on_demand_ids += [node_id]

            if on_demand_ids:
                # todo: show node names?
                cli_logger.print(
                    "Stopping instances {} " + cf.dimmed(
                        "(to terminate instead, "
                        "set `cache_stopped_nodes: False` "
                        "under `provider` in the cluster configuration)"),
                    cli_logger.render_list(on_demand_ids))

            if spot_ids:
                cli_logger.print(
                    "Terminating instances {} " +
                    cf.dimmed("(cannot stop spot instances, only terminate)"),
                    cli_logger.render_list(spot_ids))

            nodes_to_terminate[stop_instances_func] = on_demand_ids
            nodes_to_terminate[terminate_instances_func] = spot_ids
        else:
            nodes_to_terminate[terminate_instances_func] = node_ids

        max_terminate_nodes = self.max_terminate_nodes if \
            self.max_terminate_nodes is not None else len(node_ids)

        for terminate_func, nodes in nodes_to_terminate.items():
            for start in range(0, len(nodes), max_terminate_nodes):
                terminate_func(InstanceIds=nodes[start:start +
                                                 max_terminate_nodes])

    def _get_node(self, node_id):
        """Refresh and get info for this node, updating the cache."""
        self.non_terminated_nodes({})  # Side effect: updates cache

        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]

        # Node not in {pending, running} -- retry with a point query. This
        # usually means the node was recently preempted or terminated.
        matches = list(self.ec2.instances.filter(InstanceIds=[node_id]))
        assert len(matches) == 1, "Invalid instance id {}".format(node_id)
        return matches[0]

    def _get_cached_node(self, node_id):
        """Return node info from cache if possible, otherwise fetches it."""
        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]

        return self._get_node(node_id)

    def prepare_config_for_head(
            self, cluster_config: Dict[str, Any], remote_config: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a new cluster config with custom configs for head node."""
        # Since the head will use the instance profile and role to access cloud,
        # remove the client credentials from config
        clear_aws_credentials(remote_config["provider"])
        return remote_config

    def cleanup_cluster(
            self, cluster_config: Dict[str, Any], deep: bool = False):
        """Cleanup the cluster by deleting additional resources other than the nodes.
        If deep flag is true, do a deep clean up all the resources
        """
        if deep and _is_permanent_data_volumes(self.provider_config):
            delete_cluster_volumes(self.provider_config, self.cluster_name)

    def get_default_cloud_storage(self):
        """Return the managed cloud storage if configured."""
        return get_default_aws_cloud_storage(self.provider_config)

    def get_default_cloud_database(self):
        """Return the configured cloud storage if configured."""
        return get_default_aws_cloud_database(self.provider_config)

    @staticmethod
    def bootstrap_config(cluster_config):
        return bootstrap_aws(cluster_config)

    @staticmethod
    def post_prepare(
            cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Fills out missing fields after the user config is merged with defaults and before validate"""
        return post_prepare_aws(cluster_config)

    @staticmethod
    def validate_config(
            provider_config: Dict[str, Any]) -> None:
        config_dict = {
            "region": provider_config.get("region")}

        validate_config_dict(provider_config["type"], config_dict)

        storage_config = get_aws_s3_storage_config(provider_config)
        if storage_config is not None:
            config_dict = {
                "s3.bucket": storage_config.get("s3.bucket"),
                # The access key is no longer a must since we have role access
                # "s3.access.key.id": storage_config.get("s3.access.key.id"),
                # "s3.secret.access.key": storage_config.get("s3.secret.access.key")
            }

            validate_config_dict(provider_config["type"], config_dict)

    @staticmethod
    def verify_config(
            provider_config: Dict[str, Any]) -> None:
        verify_cloud_storage = provider_config.get("verify_cloud_storage", True)
        cloud_storage = get_aws_s3_storage_config(provider_config)
        if verify_cloud_storage and cloud_storage is not None:
            cli_logger.verbose("Verifying S3 storage configurations...")
            verify_s3_storage(provider_config)
            cli_logger.verbose("Successfully verified S3 storage configurations.")
